import os
import re
from metaflow import current
from metaflow.decorators import StepDecorator
from metaflow.exception import MetaflowException
from metaflow.cards import Markdown, ProgressBar
from metaflow.plugins.cards.card_decorator import CardDecorator
from metaflow.decorators import _attach_decorators_to_step

from .dbt_executor import DBTExecutor, DBTProjectConfig
from ..cards.dbt_docs import DBTDocs


class CommandNotSupported(MetaflowException):
    headline = "DBT command not supported"


class MissingProfiles(MetaflowException):
    headline = "Missing DBT Profiles configuration"


class MissingStateStorage(MetaflowException):
    headline = "Missing DBT State Storage configuration"


class MissingMacroName(MetaflowException):
    headline = "Missing macro name for run-operation"


class DbtStepDecorator(StepDecorator):
    """
    Decorator to execute DBT models before a step execution begins.


    Parameters
    ----------
    command: str, optional. Default 'run'
        DBT command to execute. Default is 'run'.
        Supported commands are: run, seed, test, build and run-operation
    macro : str, optional
        name for macro to run. Required with the 'run-operation' command
    args : str, optional
        args to pass for macro. Only used with the 'run-operation' commmand.
    project_dir: str, optional
        Path to the DBT project that contains a 'dbt_project.yml'.
        If not specified, the current folder and parent folders will be tried.
    models: List[str], optional
        List of model name(s) to run. All models will be run by default if no model name is provided.
    select: List[str], optional
        List of object name(s) to run.
    exclude: List[str], optional
        List of model name(s) to exclude. Nothing is excluded by default.
    target: str, optional
        Chooses which target to load from the profiles.yml file.
        If not specified, it will use the default target from the profiles.
    profiles: Dict[str, Union[str, Dict]]
        a configuration dictionary that will be translated into a valid profiles.yml for the dbt CLI.
    """

    name = "_dbt"

    defaults = {
        "command": "run",
        "macro": None,
        "args": None,
        "project_dir": None,
        "models": None,
        "select": None,
        "exclude": None,
        "target": None,
        "profiles": None,
        "generate_docs": False,  # TODO: This could also be true by default
        #  TODO: Add way to specify adapter through decorator as well.
    }

    def __init__(self, attributes=None, statically_defined=False):
        super(DbtStepDecorator, self).__init__(attributes, statically_defined)

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        if self.attributes["profiles"] is None and not os.path.exists("./profiles.yml"):
            raise MissingProfiles(
                "You must provide profiles configuration for the DBT decorator.\n"
                "Either provide a dictionary for the 'profiles=' attribute "
                "or create a 'profiles.yml' file in the flow folder."
            )

        cmd = self.attributes["command"]

        if cmd not in ["run", "seed", "test", "build", "run-operation"]:
            raise CommandNotSupported(f"command '{cmd}' is not supported.")

        if cmd == "run-operation" and self.attributes["macro"] is None:
            raise MissingMacroName(
                "Specify the macro name to run with the 'macro=' option on the DBT decorator for the *%s* step."
                % step_name
            )

        # Do we need persisted state due to the selectors or not?
        self.use_state = self.attributes["models"] and any(
            any(sel in val for val in self.attributes["models"])
            for sel in ["result:", "state:"]
        )

        self.ds_type = flow_datastore.TYPE

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        self.run_id = run_id
        self.step_name = step_name

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        # We want to use a run and task independent prefix for the state store,
        # so that consecutive executions have a known location to look in for previous state
        # TODO: cover projects.
        # TODO: evaluate prefix for sufficient uniqueness
        state_prefix = f"{flow.name}/{step_func.__name__}"
        build_cache_prefix = f"{flow.name}/{self.run_id}"
        with DBTExecutor(
            models=self.attributes["models"],
            macro=self.attributes["macro"],
            args=self.attributes["args"],
            select=self.attributes["select"],
            exclude=self.attributes["exclude"],
            project_dir=self.attributes["project_dir"],
            target=self.attributes["target"],
            profiles=self.attributes["profiles"],
            state_prefix=state_prefix if self.use_state else None,
            build_cache_prefix=build_cache_prefix,
            ds_type=self.ds_type,
        ) as executor:
            # execute the command

            cmd_exception = None
            try:
                for output in executor.command(self.attributes["command"]):
                    print(output, end="")
                    self._parse_result_card(output)
            except Exception as ex:
                # we do not want to immediately fail in case of a failing DBT command,
                # as we still want to try and persist artifacts etc.
                cmd_exception = ex

            if self.attributes["generate_docs"]:
                try:
                    # This might fail due to DBT version not supporting docs creation.
                    # We don't want to fail outright due to docs alone
                    out = ""
                    for line in executor.command("docs"):
                        out += line
                except Exception:
                    print(out)
                    pass

            # Write DBT run artifacts as task artifacts.
            # TODO: don't hardcode artifacts if at all not necessary.
            def _dbt_artifacts_iterable():
                artifacts = {
                    "run_results": executor.run_results,
                    "semantic_manifest": executor.semantic_manifest,
                    "manifest": executor.manifest,
                    "sources": executor.sources,
                    "catalog": executor.catalog,
                    "static_index": executor.static_index,
                }
                for name, func in artifacts.items():
                    val = func()
                    if val is None:
                        continue
                    yield (name, val)

            for k, v in _dbt_artifacts_iterable():
                setattr(flow, k, v)

            # update possible errors to the results card
            self._add_errors(executor.run_results())

            if cmd_exception:
                # finally raise if an error was encountered with the DBT command earlier.
                raise cmd_exception

        return step_func

    def _parse_result_card(self, output):
        # Init card if missing
        if not getattr(self, "result_card", None):
            self.result_card = current.card["dbt_results"]

        if not getattr(self, "result_header", None):
            self.result_header = Markdown("# Running DBT Command")
            self.result_card.append(self.result_header)

        if not getattr(self, "result_progress", None):
            self.result_progress = None
        # DBT logs progress in stdout with 'x of y' pattern for steps, so we rely on this for parsing.
        # example:
        # 1 of 50 START
        # 1 of 50 OK
        progress_info_regex = re.compile(r".* (\d+) of (\d+) OK.*")

        matched = progress_info_regex.match(output)
        if matched:
            cur, total = matched.group(1, 2)
            if not self.result_progress:
                self.result_progress = ProgressBar(max=total, label="Parts done")
                self.result_card.append(self.result_progress)

            self.result_progress.update(cur)
            self.result_card.refresh()

    def _add_errors(self, run_results=None):
        if not run_results:
            return
        # Init card if missing
        if not getattr(self, "result_card", None):
            self.result_card = current.card["dbt_results"]

        failures = [
            result for result in run_results["results"] if result["status"] != "success"
        ]
        for failure in failures:
            failure_comp = [
                Markdown("# Error with: %s" % failure["unique_id"]),
                Markdown(failure["message"]),
            ]
            self.result_card.extend(failure_comp)

        self.result_card.refresh()

    def add_to_package(self):
        """
        Called to add custom packages needed for a decorator. This hook will be
        called in the `MetaflowPackage` class where metaflow compiles the code package
        tarball. This hook is invoked in the `MetaflowPackage`'s `path_tuples`
        function. The `path_tuples` function is a generator that yields a tuple of
        `(file_path, arcname)`.`file_path` is the path of the file in the local file system;
        the `arcname` is the path of the file in the constructed tarball or the path of the file
        after decompressing the tarball.

        Returns a list of tuples where each tuple represents (file_path, arcname)
        """
        config = DBTProjectConfig(self.attributes["project_dir"])
        paths = config.project_file_paths()

        # TODO: verify keys for possible collisions.
        files = [(path, path) for path in paths]
        return files



class dbt_deco:
    def __init__(self, **kwargs):
        self.generate_docs = kwargs.get("generate_docs", False)
        self.kwargs = kwargs

    def __call__(self, step_func):
        # Make generated docs available as a card.
        from metaflow import _dbt, card

        dbt_card = card(type="blank", refresh_interval=1, id="dbt_results")
        if self.generate_docs:

            return dbt_card(card(type="dbt_docs")(_dbt(**self.kwargs)(step_func)))
        else:
            return dbt_card(_dbt(**self.kwargs)(step_func))



from copy import deepcopy
from itertools import chain
from typing import List

PARSE_TYPES = ("raw", "leaves")


class AdditionalDeco:
    def __init__(self, name, attr: List[str] = [], constant: str = None) -> None:
        self.name = name
        self.attr = attr
        self.constant = constant

    @property
    def decospec(self):
        return f"@{self.name}({self.attr})" if self.attr else f"@{self.name}"


class FlowGenerator:
    def __init__(
        self,
        manifest,
        flow_name,
        project_dir,
        flow_decorators: List[AdditionalDeco] = [],
        step_decorators: List[AdditionalDeco] = [],
        parse_type="raw",
    ):
        self.name = flow_name
        self.manifest = manifest
        self.flow_str = None
        self.parse_type = parse_type
        self.extra_constants = [
            deco.constant
            for deco in (flow_decorators + step_decorators)
            if deco.constant is not None
        ]
        self.extra_flow_decorators = [deco.decospec for deco in flow_decorators]
        self.extra_step_decorators = [deco.decospec for deco in step_decorators]
        self.extra_imports = [deco.name for deco in (flow_decorators + step_decorators)]
        # we're not interested in the default case, as we do not need to specify this in the decorator
        # attributes.
        self.project_dir = project_dir if not project_dir == "./" else None

    def generate(self):
        # crawl through
        # - manifest['child_map'] : lists children of a node
        # - manifest['parent_map'] : lists dependencies for a node
        # to form a graph for the Flow
        child_map = deepcopy(self.manifest["child_map"])
        parent_map = deepcopy(self.manifest["parent_map"])

        # Sort the nodes based on dependencies
        def _visit(nodes: set, sorted: List):
            if not nodes:
                return sorted
            _nodes = set(nodes)
            for node in nodes:
                if not parent_map[node]:
                    sorted.append(node)
                    _nodes.remove(node)
                    for child in child_map[node]:
                        parent_map[child].remove(node)
            return _visit(_nodes, sorted)

        sorted_nodes = _visit(set(child_map.keys()), [])

        supported_sorted_nodes = list(
            filter(
                lambda node: node.startswith("model") or node.startswith("seed"),
                sorted_nodes,
            )
        )
        parsers = {"raw": self.raw_parsing, "leaves": self.leaf_parsing}

        node_objs = parsers[self.parse_type](supported_sorted_nodes)

        self.flow_str = self.flow_template(node_objs)

    def raw_parsing(self, sorted_nodes):
        # Use the list of sorted nodes to construct steps
        node_objs = []
        for i, node in enumerate(sorted_nodes):
            next_node = sorted_nodes[i + 1] if i < len(sorted_nodes) - 1 else None
            cmd = node.split(".", 1)[0]
            cmd = None if cmd == "model" else cmd
            node_objs.append(
                DBTNode(
                    self.manifest["nodes"][node]["name"],
                    self.manifest["nodes"][next_node]["name"] if next_node else None,
                    cmd,
                    self.project_dir is not None,
                    self.extra_step_decorators,
                )
            )
        return node_objs

    def leaf_parsing(self, sorted_nodes):
        # filter leaves from nodes
        parent_map = self.manifest["parent_map"]
        child_map = self.manifest["child_map"]
        leaves = []
        for node in sorted_nodes:
            true_children = [n for n in child_map[node] if not n.startswith("test.")]
            if not true_children and parent_map[node]:
                leaves.append(node)

        # Seeds need to be run separately.
        seeds_and_leaves = [
            n for n in sorted_nodes if n.startswith("seed.") or n in leaves
        ]

        def _gather(node):
            # gather dependencies for a node recursively.
            deps = []
            parents = [
                p
                for p in parent_map[node]
                if p.startswith("seed.") or p.startswith("model.")
            ]
            if parents:
                deps.extend(parents)
                for parent in parents:
                    deps.extend(_gather(parent))
            return deps

        # construct tree with branches
        tree = []
        tail = [
            n for n in sorted_nodes if n.startswith("seed.") or n.startswith("model.")
        ]
        for n in seeds_and_leaves:
            branch = [n for n in [n] + _gather(n)]
            tree.append([n for n in sorted_nodes if n in branch])
            tail = [n for n in tail if n not in branch]

        # add tail branch if one exists
        tree.append([n for n in sorted_nodes if n in tail])

        def shared_ancestors_set(branches=[]):
            # return a set of nodes that are shared by at least one branch (and need to therefore be deduplicated)
            common_ancestors = set()
            for branch in branches:
                for n in branch:
                    if (
                        n not in common_ancestors
                        and len([n for b in branches if n in b]) > 1
                    ):
                        common_ancestors.add(n)

            return common_ancestors

        def partition(branch):
            # partition a branch into subsets of similar type of nodes, up to a maximum number of nodes
            ntype = None
            part = []
            maxnodes = 10
            sorted_partitions = []
            for n in branch:
                nt, _ = n.split(".", 1)
                if ntype is None:
                    ntype = nt
                if nt != ntype:
                    # save current collected nodes
                    sorted_partitions.append(part)
                    ntype = nt
                    # start a new partition
                    part = [n]
                else:
                    part.append(n)
            # append last part outside of the iteration
            sorted_partitions.append(part)

            return sorted_partitions

        def parse_tree(branches):
            ancestry = shared_ancestors_set(branches)
            sorted_ancestry = [n for n in sorted_nodes if n in ancestry]
            trimmed_branches = [[n for n in b if n not in ancestry] for b in branches]
            # remove empty branches
            trimmed_branches = [branch for branch in trimmed_branches if branch]

            return sorted_ancestry, trimmed_branches

        ancestors, branches = parse_tree(tree)

        # Use the list of sorted nodes to construct steps
        node_objs = []

        branch_entry_fns = []
        for branch in branches:
            branch_name_stub = self.get_node_name(branch[-1])
            pt = partition(branch)
            for idx, part in enumerate(pt):
                name = "dbt_%s_%d" % (branch_name_stub, idx)
                if idx == 0:
                    branch_entry_fns.append(name)
                nextname = (
                    "dbt_%s_%d" % (branch_name_stub, idx + 1)
                    if idx < len(pt) - 1
                    else "join_dbt_branches"
                )
                cmd = get_node_cmd(part[0])
                node_objs.append(
                    DBTMultiNode(
                        name,
                        [nextname],
                        cmd,
                        self.project_dir is not None,
                        self.extra_step_decorators,
                        [self.get_node_name(n) for n in part],
                    )
                )
        node_objs.append(JOIN_NODE)

        ancestor_nodes = []
        ancestor_parts = partition(ancestors)
        for idx, part in enumerate(ancestor_parts):
            name = "dbt_shared_parents_%d" % idx
            nextnames = (
                ["dbt_shared_parents_%d" % (idx + 1)]
                if idx < len(ancestor_parts) - 1
                else branch_entry_fns
            )
            cmd = get_node_cmd(part[0])
            ancestor_nodes.append(
                DBTMultiNode(
                    name,
                    nextnames,
                    cmd,
                    self.project_dir is not None,
                    self.extra_step_decorators,
                    [self.get_node_name(n) for n in part],
                )
            )

        return ancestor_nodes + node_objs

    def get_node_name(self, node):
        return self.manifest["nodes"][node]["name"]

    def write(self):
        if not self.flow_str:
            raise Exception("Flow not generated, nothing to write yet.")
        filename = str(self.name).lower() + ".py"
        with open(filename, "w") as f:
            f.write(self.flow_str)

        return filename

    def flow_template(self, dbt_nodes: List):
        first_node_name = dbt_nodes[0].name
        constants = "".join(
            const
            for const in [
                f'PROJECT_DIR="{self.project_dir}"' if self.project_dir else None,
            ]
            + [
                f"""
{const}"""
                for const in self.extra_constants
            ]
            if const is not None
        )
        metaflow_imports = ", ".join(
            [
                dep
                for dep in [
                    "step",
                    "FlowSpec",
                    "dbt",
                ]
                if dep is not None
            ]
            + self.extra_imports
        )

        flow_decos = "".join(
            f"""
{deco}"""
            for deco in self.extra_flow_decorators
        )
        template = f"""
from metaflow import {metaflow_imports}

{constants}
{flow_decos}
class {self.name}(FlowSpec):
    
    @step
    def start(self):
        print("Starting the generated DBT Flow")
        self.next(self.{sanitize(first_node_name)})"""
        for node in dbt_nodes:
            template += str(node)

        template += f"""
    @step
    def end(self):
        print("DBT executed successfully")

if __name__=="__main__":
    {self.name}()
"""

        return template


class DBTNode:
    def __init__(
        self,
        node_name: str,
        next_node_name: str = None,
        command: str = None,
        project_dir=False,
        extra_step_decorators=[],
        include_parents=False,
    ):
        # necessary as model names might start with a number, but python functions cannot.
        func_prefix = "dbt_"
        self.name = func_prefix + node_name
        self.next_node_name = (
            (func_prefix + next_node_name) if next_node_name else "end"
        )
        include_op = "+" if include_parents else ""
        self.cmd_str = command if command else "run"
        deco_attrs = ", ".join(
            attr
            for attr in [
                f'command="{command}"' if command else None,
                f'models=["{include_op}{node_name}"]',
                "project_dir=PROJECT_DIR" if project_dir else None,
            ]
            if attr is not None
        )
        self.deco_stack = "".join(
            deco
            for deco in [
                f"""
    @dbt({deco_attrs})""",
            ]
            + [
                f"""
    {deco}"""
                for deco in extra_step_decorators
            ]
            if deco is not None
        )

    def __str__(self) -> str:
        return f"""
    {self.deco_stack}
    @step
    def {sanitize(self.name)}(self):
        self.next(self.{sanitize(self.next_node_name)})
"""


class DBTMultiNode:
    def __init__(
        self,
        func_name: str,
        next_funcs: List[str] = None,
        command: str = None,
        project_dir=False,
        extra_step_decorators=[],
        models=[],
    ):
        self.func_name = func_name
        self.name = func_name

        self.next_func_str = ", ".join(f"self.{sanitize(fn)}" for fn in next_funcs)
        self.cmd_str = command if command else "run"
        model_list = ", ".join(f'"{name}"' for name in models)
        deco_attrs = ", ".join(
            attr
            for attr in [
                f'command="{command}"' if command else None,
                f"models=[{model_list}]",
                "project_dir=PROJECT_DIR" if project_dir else None,
            ]
            if attr is not None
        )
        self.deco_stack = "".join(
            deco
            for deco in [
                f"""
    @dbt({deco_attrs})""",
            ]
            + [
                f"""
    {deco}"""
                for deco in extra_step_decorators
            ]
            if deco is not None
        )

    def __str__(self) -> str:
        return f"""
    {self.deco_stack}
    @step
    def {sanitize(self.func_name)}(self):
        self.next({self.next_func_str})
"""


JOIN_NODE = """
    @step
    def join_dbt_branches(self, inputs):
        self.next(self.end)
"""


def sanitize(name: str):
    return name.replace(".", "_")


def get_node_cmd(node):
    return "seed" if node.startswith("seed.") else None



from metaflow.metaflow_config_funcs import from_conf

###
# CONFIGURE: You can also specify additional debugging options that are
#            activated using the METAFLOW_DEBUG_XXX environment variable. See debug.py
###
DEBUG_OPTIONS = ["dbt"]



__mf_extensions__ = "dbt-ext"

# Make the switch decorator available at the top level.
from ..plugins.dbt import dbt_deco as dbt

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("metaflow-dbt").version
except:
    # this happens on remote environments since the job package
    # does not have a version
    __version__ = None



from typing import Dict, List
from metaflow._vendor import click
from .flow_generator import FlowGenerator, PARSE_TYPES, AdditionalDeco
import json

GENERATE_META = ".generate_config"


@click.group()
@click.pass_context
def cli(ctx):
    pass


@cli.group(help="Commands for DBT on Metaflow")
def dbt():
    pass


@dbt.command(help="Generate a Flow from an existing DBT project")
@click.option(
    "--ci",
    default=False,
    type=bool,
    is_flag=True,
    help="CI mode, which generates based on predefined defaults and values stored in '.generate_config'",
)
@click.pass_obj
def generate(obj, ci=False):
    conf = {
        "project_dir": "./",
        "flow_name": "DBTFlow",
        "remote_config": False,
        "flow_decorators": {
            # TODO: make this configurable
            "conda_base": {"attr": 'packages={"dbt-postgres": ">=1.7.0"}'}
            # "trigger_event": False
        },
        "step_decorators": {
            # "secrets": False,
            # "environment": False,
            # "retry": False
        },
        "parse_type": "raw",
    }
    try:
        with open(GENERATE_META, "r") as f:
            conf = {**conf, **json.load(f)}
    except Exception:
        pass  # cached config load error, not important.

    # Prompt user for configuration if not in CI mode.
    if not ci:
        conf = prompt_config_options(conf)

        # record answers for conf
        with open(GENERATE_META, "w+") as f:
            json.dump(conf, f)

    from ..plugins.dbt.dbt_executor import DBTExecutor

    with DBTExecutor(project_dir=conf["project_dir"]) as executor:
        click.secho("Parsing DBT Project...", bold=True)
        executor.parse()

        click.secho("Generating Flow %s..." % conf["flow_name"], bold=True)

        manifest = executor.manifest()
        generator = FlowGenerator(
            manifest,
            conf["flow_name"],
            conf["project_dir"],
            [
                AdditionalDeco(name=name, **args)
                for name, args in conf["flow_decorators"].items()
                if args
            ],
            [
                AdditionalDeco(name=name, **args)
                for name, args in conf["step_decorators"].items()
                if args
            ],
            conf["parse_type"],
        )
        generator.generate()

        filename = generator.write()
        click.secho(f"Flow has been saved as {filename}")


def prompt_config_options(conf):
    conf["project_dir"] = click.prompt(
        "Path to the DBT project: ", default=conf["project_dir"], show_default=True
    )
    conf["flow_name"] = click.prompt(
        "Name for generated Flow: ", default=conf["flow_name"], show_default=True
    )

    # Configure decorators

    # RETRY
    if click.confirm(
        "Do you want to retry DBT steps in case it fails?",
        default=bool(conf["step_decorators"].get("retry", False)),
    ):
        retries = click.prompt("How many times do you want to retry?", type=int)
        conf["step_decorators"]["retry"] = {"attr": f"times={retries}"}

    # Config for remote execution environments.
    conf["remote_config"] = click.confirm(
        "Do you need to supply values to your DBT Profile?",
        default=conf["remote_config"],
    )
    if conf["remote_config"]:
        # SECRETS
        if click.confirm(
            "Read values from a secrets provider? (Suggested for sensitive information, such as database credentials)",
            default=conf["step_decorators"].get("secrets", False),
        ):
            secret_key = click.prompt(
                "What is the name of the secret that you want to use?", type=str
            )
            conf["step_decorators"]["secrets"] = {
                "attr": "sources=SECRET_SRC",
                "constant": f'SECRET_SRC=["{secret_key}"]',
            }

        # ENVIRONMENT
        if click.confirm(
            "Supply values as environment variables?",
            default=conf["step_decorators"].get("environment"),
        ):
            inputs = keep_prompting(
                "Enter a value for the environment in the form of VARIABLE='some value'"
            )
            envs = dict(ip.split("=") for ip in inputs)
            conf["step_decorators"]["environment"] = {
                "attr": "vars=ENVS",
                "constant": f"ENVS={envs}",
            }

    else:
        # reset values as they might be true from previous generate.
        conf["step_decorators"].pop("secrets")
        conf["step_decorators"].pop("environment")

    conf["parse_type"] = click.prompt(
        "How should we parse the project?", type=click.Choice(PARSE_TYPES)
    )

    return conf


def keep_prompting(prompt, acc=[]):
    click.echo(acc)
    val = click.prompt(f"{prompt} (Empty to confirm.)", default="", show_default=False)
    if val == "":
        return acc
    acc.append(val)
    return keep_prompting(prompt, acc)



STEP_DECORATORS_DESC = [("_dbt", ".dbt.dbt_decorator.DbtStepDecorator")]

###
# CONFIGURE: Similar to datatools, you can make visible under metaflow.plugins.* other
#            submodules not referenced in this file
###
__mf_promote_submodules__ = []



from setuptools import setup, find_namespace_packages

version = "1.0.4"

setup(
    name="metaflow-dbt",
    version=version,
    description="DBT extension for Metaflow",
    long_description=open("README.md").read(),
    author="Outerbounds",
    author_email="help@outerbounds.co",
    long_description_content_type="text/markdown",
    license="Apache Software License",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    packages=find_namespace_packages(include=["metaflow_extensions.*"]),
    py_modules=[
        "metaflow_extensions",
    ],
    python_requires=">=3.8",
    install_requires=["dbt-core>=1.7.0", "pyyaml"],
)



from metaflow.cards import MetaflowCard, BlankCard
from metaflow.plugins.cards.card_modules.basic import (
    SectionComponent,
    TitleComponent,
    TextComponent,
)


class DBTDocs(MetaflowCard):
    type = "dbt_docs"

    def render(self, task):  # this function returns the HTML to be rendered
        return task.data.static_index


CARDS = [DBTDocs]



import subprocess
import os
import tempfile
import json
import yaml
import glob
import shutil
from typing import Dict, List, Optional
import tarfile

from metaflow.exception import MetaflowException
from metaflow.util import which
from metaflow.plugins.datatools.s3 import S3
from metaflow.metaflow_config import DEBUG_DBT


class DBTExecutionFailed(MetaflowException):
    headline = "DBT Run execution failed"


class UnsupportedCommand(MetaflowException):
    headline = "Command is not supported"


class DBTDuplicateDependencyFile(MetaflowException):
    headline = "Duplicate dependency file found"


PRECOMPILE_METAFILE = ".precompile.done"
STATE_PUSH_METAFILE = ".state-push.done"
STATE_PULL_METAFILE = ".state-pull.done"


def debug_log(line: str):
    if not DEBUG_DBT:
        return
    print(line)


# TODO: There is a choice to utilize the Python library provided by DBT for performing the run, and accessing run results as well.
# This would introduce a heavy dependency for the decorator use case, which can be completely avoided with the custom implementation
# via calling the CLI via subprocess only at the point when execution needs to happen. Decide on the approach after PoC is complete.
class DBTExecutor:
    def __init__(
        self,
        models: List[str] = None,
        macro: str = None,
        args: str = None,
        select: List[str] = None,
        exclude: List[str] = None,
        project_dir: str = None,
        target: str = None,
        profiles: Dict = None,
        state_prefix: str = None,
        build_cache_prefix: str = None,
        ds_type=None,
    ):

        self.models = " ".join(models) if models is not None else None
        self.select = " ".join(select) if select is not None else None
        self.exclude = " ".join(exclude) if exclude is not None else None
        self.macro = macro
        self.args = args
        self.project_dir = project_dir
        self.target = target
        self.bin = which("./dbt") or which("dbt")
        if self.bin is None:
            raise DBTExecutionFailed("Can not find DBT binary. Please install DBT")
        self.profiles = profiles
        conf = DBTProjectConfig(project_dir)
        self._project_config = conf.project_config

        _ = (
            conf.project_file_paths()
        )  # trigger setting the dbt_dependencies member variable
        self._dbt_dependencies = conf.dbt_dependencies

        self.state_store = None
        if state_prefix:
            self.state_store = self._init_datastore(ds_type, state_prefix)
        self.build_cache = None
        if build_cache_prefix:
            # we want to add the dbt project name to the cache prefix to avoid collisions,
            # as a flow run can be executing multiple dbt projects
            project_name = self._project_config.get("name", None)
            prefix = (
                "%s/%s" % (build_cache_prefix, project_name)
                if project_name
                else build_cache_prefix
            )
            self.build_cache = self._init_datastore(ds_type, prefix)
        self.tempdir = None
        self.in_context = False

    @staticmethod
    def _init_datastore(ds_type, prefix: str = ""):
        from metaflow.plugins import DATASTORES

        datastore = [d for d in DATASTORES if d.TYPE == ds_type][0]

        root = datastore.get_datastore_root_from_config(print)
        return datastore(f"{root}/dbt_state/{prefix}")

    def run_results(self) -> Optional[Dict]:
        return self._read_dbt_artifact("run_results.json")

    def semantic_manifest(self) -> Optional[Dict]:
        return self._read_dbt_artifact("semantic_manifest.json")

    def manifest(self) -> Optional[Dict]:
        return self._read_dbt_artifact("manifest.json")

    def catalog(self) -> Optional[Dict]:
        return self._read_dbt_artifact("catalog.json")

    def sources(self) -> Optional[Dict]:
        return self._read_dbt_artifact("sources.json")

    def static_index(self) -> Optional[Dict]:
        return self._read_dbt_artifact("static_index.html", raw=True)

    def command(self, cmd: str):
        supported = {
            "run": self.run,
            "parse": self.parse,
            "seed": self.seed,
            "build": self.build,
            "test": self.test,
            "run-operation": self.run_operation,
            "docs": self.generate_docs,
        }
        if cmd not in supported:
            raise UnsupportedCommand(
                'The command "%s" is not supported by the DBT Extension' % cmd
            )

        return supported[cmd]()

    def run(self) -> str:
        args = ["--fail-fast", "--no-use-colors"]
        if self.project_dir is not None:
            args.extend(["--project-dir", self.project_dir])
        if self.models is not None:
            args.extend(["--models", self.models])
        if self.select is not None:
            args.extend(["--select", self.select])
        if self.exclude is not None:
            args.extend(["--exclude", self.exclude])
        if self.target is not None:
            args.extend(["--target", self.target])

        return self._call("run", args)

    def test(self) -> str:
        args = ["--fail-fast", "--no-use-colors"]
        if self.project_dir is not None:
            args.extend(["--project-dir", self.project_dir])
        if self.models is not None:
            args.extend(["--models", self.models])
        if self.select is not None:
            args.extend(["--select", self.select])
        if self.target is not None:
            args.extend(["--target", self.target])

        return self._call("test", args)

    def build(self) -> str:
        args = ["--fail-fast", "--no-use-colors"]
        if self.project_dir is not None:
            args.extend(["--project-dir", self.project_dir])
        if self.models is not None:
            args.extend(["--models", self.models])
        if self.select is not None:
            args.extend(["--select", self.select])
        if self.target is not None:
            args.extend(["--target", self.target])

        return self._call("build", args)

    def run_operation(self) -> str:
        if self.macro is None:
            raise DBTExecutionFailed(
                "run-operation requires specifying the name of a macro to be executed."
            )
        args = [self.macro, "--fail-fast", "--no-use-colors"]
        if self.args is not None:
            args.extend(["--args", self.args])
        if self.project_dir is not None:
            args.extend(["--project-dir", self.project_dir])
        if self.target is not None:
            args.extend(["--target", self.target])

        return self._call("run-operation", args)

    def parse(self) -> str:
        args = ["--fail-fast", "--no-use-colors"]
        if self.project_dir is not None:
            args.extend(["--project-dir", self.project_dir])
        if self.models is not None:
            args.extend(["--models", self.models])
        if self.select is not None:
            args.extend(["--select", self.select])
        if self.exclude is not None:
            args.extend(["--exclude", self.exclude])

        return self._call("parse", args)

    def seed(self) -> str:
        args = ["--no-use-colors"]
        if self.project_dir is not None:
            args.extend(["--project-dir", self.project_dir])
        if self.models is not None:
            args.extend(["--models", self.models])
        if self.select is not None:
            args.extend(["--select", self.select])
        if self.exclude is not None:
            args.extend(["--exclude", self.exclude])
        if self.target is not None:
            args.extend(["--target", self.target])

        return self._call("seed", args)

    def generate_docs(self) -> str:
        # The static docs generation requires dbt-core >= 1.7
        args = ["generate", "--static", "--no-compile", "--no-use-colors"]
        if self.project_dir is not None:
            args.extend(["--project-dir", self.project_dir])
        if self.models is not None:
            args.extend(["--models", self.models])
        if self.select is not None:
            args.extend(["--select", self.select])
        if self.exclude is not None:
            args.extend(["--exclude", self.exclude])
        if self.target is not None:
            args.extend(["--target", self.target])

        return self._call("docs", args)

    def _read_dbt_artifact(self, name: str, raw: bool = False):
        artifact = os.path.join(
            ".",
            self.project_dir or "",
            self._project_config.get("target-path", "target"),
            name,
        )
        try:
            with open(artifact) as m:
                return m.read() if raw else json.load(m)
        except FileNotFoundError:
            return None

    def _push_state(self):
        # Push new state to self.state_store if configured
        if not self.state_store:
            return
        target_path = os.path.join(
            ".",
            self.project_dir or "",
            self._project_config.get("target-path", "target"),
        )
        marker = os.path.join(target_path, STATE_PUSH_METAFILE)
        if os.path.exists(marker):
            debug_log("Skipped: State assets have already been pushed once.")
            return

        files_and_paths = {
            key: os.path.join(
                target_path,
                key,
            )
            for key in ["manifest.json", "run_results.json"]
        }
        files_and_handles = {
            key: open(path, mode="rb")
            for key, path in files_and_paths.items()
            if os.path.exists(path)
        }

        self.state_store.save_bytes(
            (name, content) for name, content in files_and_handles.items()
        )

        # write a marker that state has been pushed once.
        with open(marker, "w") as f:
            f.write("")

    def _pull_state(self):
        # Fetch previous state to tempdir from self.state_store if configured
        if not self.state_store:
            return None
        state_path = os.path.join(self.tempdir, "prev_state")
        os.makedirs(state_path, exist_ok=True)

        marker = os.path.join(state_path, STATE_PULL_METAFILE)
        if os.path.exists(marker):
            debug_log("Skipped: State assets have already been fetched")
            return state_path

        with self.state_store.load_bytes(
            ["manifest.json", "run_results.json"]
        ) as result:
            for key, file, _ in result:
                if file is not None:
                    shutil.move(file, os.path.join(state_path, key))

        # write a marker that state has been pulled once.
        if os.listdir(state_path):
            with open(marker, "w") as f:
                f.write("")
            return state_path
        return None

    def _pull_prebuild(self):
        if not self.build_cache:
            return
        target_path = os.path.join(
            ".",
            self.project_dir or "",
            self._project_config.get("target-path", "target"),
        )

        marker = os.path.join(target_path, PRECOMPILE_METAFILE)
        if os.path.exists(marker):
            debug_log("Skipped: Prebuild assets have already been fetched")
            return
        with self.build_cache.load_bytes(["prebuild_tar"]) as result:
            for key, file, _ in result:
                if file is None:
                    debug_log("Failed to pull prebuild assets, prebuild_tar is missing")
                    return
                with tarfile.open(file, "r:gz") as t:
                    t.extractall()
                    debug_log("Extracted contents of %s to %s " % (file, target_path))

        # Add marker that prebuild assets have been fetched
        # target_path might not exist if we did not get anything to extract from the cache.
        os.makedirs(target_path, exist_ok=True)
        with open(marker, "w") as f:
            f.write("")

    def _push_prebuild(self):
        if not self.build_cache:
            return

        target_path = os.path.join(
            ".",
            self.project_dir or "",
            self._project_config.get("target-path", "target"),
        )
        if not os.path.exists(target_path):
            return

        marker = os.path.join(target_path, PRECOMPILE_METAFILE)
        if os.path.exists(marker):
            debug_log("Skipped: Prebuild assets have already been pushed once.")
            return

        temp_tar = os.path.join(self.tempdir, "prebuild_tar")
        with tarfile.open(temp_tar, "w:gz") as t:
            t.add(target_path)
            debug_log(
                "Added assets from path %s to precompiled assets archive" % target_path
            )

        with open(temp_tar, "rb") as f:
            self.build_cache.save_bytes([("prebuild_tar", f)])
            debug_log("Persisting precompiled assets tar from %s" % temp_tar)

        # Add marker that prebuild assets have been pushed
        with open(marker, "w") as f:
            f.write("")

    def __enter__(self):
        self.tempdir_handle = tempfile.TemporaryDirectory()
        self.tempdir = self.tempdir_handle.__enter__()
        self.in_context = True
        return self

    def __exit__(self, exc, value, tb):
        self.tempdir_handle.__exit__(exc, value, tb)
        self.tempdir = None
        self.in_context = False
        return

    def persist_temp_profiles(self):
        "Write a profiles.yml to the tempdir if necessary. Returns a bool for whether a temp profile exists."
        if self.profiles is None:
            return False
        profiles_path = os.path.join(self.tempdir, "profiles.yml")
        if not os.path.exists(profiles_path):
            # Synthesize a profiles.yml from the passed in config dictionary.
            with open(profiles_path, "w") as f:
                conf = yaml.dump(self.profiles)
                f.write(conf)
                debug_log("Wrote a temporary profile to: %s" % profiles_path)
        return True

    def _call(self, cmd, args):
        if not self.in_context:
            raise Exception("DBT Command need to be called within a context manager")
        profile_args = (
            ["--profiles-dir", self.tempdir] if self.persist_temp_profiles() else []
        )

        state_args = []
        # Try using DBT state
        state_path = self._pull_state()
        if state_path:
            debug_log(f"got state assets!")
            # a previous state was available.
            state_args = ["--state", state_path]
        else:
            # we do not have a previous state,
            # so we need to clean up any known state selectors from args in order to avoid errors.
            def _cleanup(arg: str):
                split = arg.split(",")
                args = [a for a in split if not "state:" in a and not "result:" in a]
                if len(args) < len(split):
                    debug_log(
                        "cleaned up a state selector because no previous state was found."
                    )
                return ",".join(args)

            args = [_cleanup(arg) for arg in args]

        # Pull prebuild assets before DBT Command
        self._pull_prebuild()

        # Install dbt dependencies if the packages.yml or dependenices.yml is provided in the project directory.
        if (
            self._dbt_dependencies
        ):  # Assume there is no dbt_packages directory yet, and install dependencies on the fly.
            process = subprocess.Popen(
                [self.bin, "deps", "--project-dir", self.project_dir],
                stdout=subprocess.PIPE,
            )
            while True:
                process.poll()
                if process.returncode is None:
                    # process is still running
                    line = process.stdout.readline()
                    if not line:
                        # end of stdout, but process has not ended yet.
                        continue
                    yield line.decode()
                elif process.returncode == 0:
                    break
                elif process.returncode != 0:
                    raise DBTExecutionFailed(
                        "Ran into an issue with dbt command: dbt deps."
                    )

        try:
            process = subprocess.Popen(
                [self.bin, cmd] + args + profile_args + state_args,
                stdout=subprocess.PIPE,
            )
            while True:
                process.poll()
                if process.returncode is None:
                    # process is still running
                    line = process.stdout.readline()
                    if not line:
                        # end of stdout, but process has not ended yet.
                        continue
                    yield line.decode()
                elif process.returncode == 0:
                    break
                elif process.returncode != 0:
                    # TODO: would be a nice addition if stdout and stderr were separated here.
                    raise DBTExecutionFailed(
                        "ran into an issue with dbt command: %s" % cmd
                    )
        finally:
            # Push state artifacts to self.state_store
            self._push_state()
            self._push_prebuild()


# We want a separate construct for the project config, so this can be parsed without requiring the dbt binary to be present on the system.
# This way users deploying to remote execution do not need to install DBT on their own machine.
class DBTProjectConfig:
    def __init__(self, project_dir: str = None):
        self.project_dir = project_dir

    @property
    def project_config(self):
        config_path = os.path.join(self.project_dir or "./", "dbt_project.yml")

        try:
            with open(config_path) as f:
                return yaml.load(f, Loader=yaml.Loader)
        except FileNotFoundError:
            raise MetaflowException("No configuration file 'dbt_project.yml' found")

    def project_file_paths(self):
        """
        Return a list of files required for the DBT project.
        Used to include necessary files in the codepackage
        """
        files = []
        _inc = [
            f
            for f in [
                "./profiles.yml",
                os.path.join(self.project_dir or "", "dbt_project.yml"),
            ]
            if os.path.exists(f)
        ]
        files.extend(_inc)

        for component in [
            "model-paths",
            "seed-paths",
            "test-paths",
            "analysis-paths",
            "macro-paths",
        ]:
            rel_path = self.project_config.get(component, [None])[
                0
            ]  # dbt profile config defines the file path inside a List.
            if rel_path is None:
                continue
            component_path = os.path.join(self.project_dir or "", rel_path)
            if not os.path.exists(component_path):
                continue
            for path in glob.glob(os.path.join(component_path, "**"), recursive=True):
                files.append(path)

        # Handle dbt packages
        # Mode 1: user has run dbt deps locally, producing the dbt_packages directory.
        # Mode 1 is for the local execution case, where the user has dbt installed locally and wants to minimize edit distance between local and remote execution.
        dbt_packages = os.path.join(self.project_dir or "", "dbt_packages")
        self.dbt_dependencies = []
        if os.path.exists(dbt_packages):
            for path in glob.glob(os.path.join(dbt_packages, "**"), recursive=True):
                files.append(path)
        # Mode 2: user has a packages.yml or dependencies.yml, include this in the codepackage to run dbt deps on the fly.
        # Mode 2 is for the remote execution case, where the user need not have dbt installed locally.
        else:
            found_files = []
            for file in ["packages.yml", "dependencies.yml"]:
                file_path = os.path.join(self.project_dir or "", file)
                if os.path.exists(file_path):
                    found_files.append(file_path)
            if len(found_files) > 1:
                raise DBTDuplicateDependencyFile(
                    "Found 'packages.yml' and 'dependencies.yml' in the dbt project directory: %s. Choose one. Read more at %s."
                    % (found_files, "https://docs.getdbt.com/docs/build/packages")
                )
            if found_files:
                self.dbt_dependencies = found_files[0]
                files.append(found_files[0])

        return files
